from unittest.mock import Mock, call

import pytest
from baby_steps import given, then, when
from d42 import ValidationException, validate_or_fail

try:
    from d42.declaration import Schema
except ImportError:
    from district42.types import Schema

from vedro.core import Dispatcher, Plugin, ScenarioResult
from vedro.events import (
    ExceptionRaisedEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioRunEvent,
)

from vedro_d42_validator import D42Validator, D42ValidatorPlugin
from vedro_d42_validator._d42_validator import validate_or_fail_ne

from ._utils import (
    dispatcher,
    format_exc_info,
    make_exc_info,
    patch_override,
    raise_nested_exception,
    validator,
)

__all__ = ("dispatcher", "validator",)  # fixtures


def test_validator():
    with when:
        validator = D42ValidatorPlugin(D42Validator)

    with then:
        assert isinstance(validator, Plugin)


@pytest.mark.usefixtures(validator.__name__)
async def test_scenario_run_event(*, dispatcher: Dispatcher):
    with given:
        scenario_result = Mock(ScenarioResult)
        event = ScenarioRunEvent(scenario_result)

    with when, patch_override() as mock:
        await dispatcher.fire(event)

    with then:
        assert mock.mock_calls == [
            call("__eq__", validate_or_fail),
            call("__ne__", validate_or_fail_ne)
        ]


@pytest.mark.usefixtures(validator.__name__)
@pytest.mark.parametrize("event_class", [ScenarioPassedEvent, ScenarioFailedEvent])
async def test_scenario_end_event(event_class, *, dispatcher: Dispatcher):
    with given:
        scenario_result = Mock(ScenarioResult)
        with patch_override():
            await dispatcher.fire(ScenarioRunEvent(scenario_result))
        event = event_class(scenario_result)

    with when, patch_override() as mock:
        await dispatcher.fire(event)

    with then:
        assert mock.mock_calls == [
            call("__eq__", Schema.__eq__),
            call("__ne__", Schema.__ne__)
        ]


@pytest.mark.usefixtures(validator.__name__)
async def test_exception_raised_event(*, dispatcher: Dispatcher):
    with given:
        exc_info = make_exc_info(AssertionError())
        formatted = format_exc_info(exc_info)
        event = ExceptionRaisedEvent(exc_info)

    with when:
        await dispatcher.fire(event)

    with then:
        assert format_exc_info(exc_info) == formatted


@pytest.mark.usefixtures(validator.__name__)
async def test_exception_raised_validation_event(*, dispatcher: Dispatcher):
    with given:
        exc_info = make_exc_info(ValidationException(), raise_nested_exception)
        formatted = format_exc_info(exc_info)
        event = ExceptionRaisedEvent(exc_info)

    with when:
        await dispatcher.fire(event)

    with then:
        assert format_exc_info(exc_info) == formatted[:4] + formatted[5:]



from typing import Any, Type, Union

from d42 import ValidationException, validate_or_fail

try:
    from d42.declaration import GenericSchema, Schema
except ImportError:
    from district42.types import Schema, GenericSchema  # type: ignore

from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.events import (
    ExceptionRaisedEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioRunEvent,
)

__all__ = ("D42Validator", "D42ValidatorPlugin",)


def validate_or_fail_ne(self: GenericSchema, other: Any) -> bool:
    try:
        validate_or_fail(self, other)
    except ValidationException:
        return True
    else:
        raise ValidationException(
            f"Value {type(other)} must not be equal to {self!r}, but {other!r} given"
        )


class D42ValidatorPlugin(Plugin):
    def __init__(self, config: Type["D42Validator"]) -> None:
        super().__init__(config)
        self._eq: Any = None
        self._ne: Any = None

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ScenarioRunEvent, self.on_scenario_run) \
                  .listen(ScenarioPassedEvent, self.on_scenario_end) \
                  .listen(ScenarioFailedEvent, self.on_scenario_end) \
                  .listen(ExceptionRaisedEvent, self.on_exception_raised)

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        self._eq = Schema.__eq__
        self._ne = Schema.__ne__
        Schema.__override__("__eq__", validate_or_fail)
        Schema.__override__("__ne__", validate_or_fail_ne)

    def on_scenario_end(self, event: Union[ScenarioPassedEvent, ScenarioFailedEvent]) -> None:
        Schema.__override__("__eq__", self._eq)
        Schema.__override__("__ne__", self._ne)

    def on_exception_raised(self, event: ExceptionRaisedEvent) -> None:
        if event.exc_info.type is not ValidationException:
            return
        prev = event.exc_info.traceback
        tb = prev.tb_next
        while tb is not None:
            if tb.tb_next is None:
                prev.tb_next = None
            prev = tb
            tb = tb.tb_next


class D42Validator(PluginConfig):
    plugin = D42ValidatorPlugin
